
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/constraints/common/constraintsparameters.hh>

#define GRID_EPS 1e-7
#define GRID_WIDTH 6      // 1.5 for summer school //2 for contactpointtest //6 for full summer school

#define GRID_BOTTOM 0   //0 For summer school   //0.35 for contactpointtest
#define GRID_TOP 0.125  //0.125 for summer school      //1 for contactpointtest

class InflowOutflow_v : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( (xg[0]> GRID_WIDTH-GRID_EPS  /*|| xg[0] < GRID_EPS*/) && xg[1]>GRID_BOTTOM+GRID_EPS && xg[1]<GRID_TOP-GRID_EPS)
    {
      return false;
    }
    return true;
  }
};



class InflowOutflow_p : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ((xg[0]> GRID_WIDTH-GRID_EPS /*|| xg[0] < GRID_EPS*/) && xg[1]>GRID_BOTTOM+GRID_EPS && xg[1]<GRID_TOP-GRID_EPS)
    {
      return true;
    }
    return false;
  }
};

class InflowOutflow_u : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if (xg[0]< GRID_EPS)
    {
      return true;
    }
    return false;
  }
};

class InflowOutflow_phi : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( (xg[0]< GRID_EPS || xg[0] >GRID_WIDTH-GRID_EPS) && xg[1]>GRID_BOTTOM+GRID_EPS && xg[1]<GRID_TOP-GRID_EPS)
    {
      return true;
    }
    return false;
  }
};
