#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>




template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
    class FEM_FFS_CHNS_NavierStokes :
      public Dune::PDELab::FullVolumePattern,
      public Dune::PDELab::LocalOperatorDefaultFlags,
      public Dune::PDELab::NumericalJacobianVolume<FEM_FFS_CHNS_NavierStokes<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
      public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FFS_CHNS_NavierStokes<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
    {
    private:
      typedef typename CHContainer::field_type RF;
      typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

      CHLocalBasisCache<CHGFS> chCache;
      VLocalBasisCache<NSGFS> vCache;
      PLocalBasisCache<NSGFS> pCache;

      typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
      mutable CHLView chLocal;
      //mutable CHLView chOldLocal;

      typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
      mutable NSLView nsOldLocal;

      Param& param; // parameter functions

      //FOR DETECTING ENTITY changes
      typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENT;
      mutable ENT OldEnt;

    public:


      // pattern assembly flags
      enum { doPatternVolume = true };

      // residual assembly flags
      enum { doAlphaVolume = true };
      //enum { doLambdaVolume = true };



      FEM_FFS_CHNS_NavierStokes (Param& param_, CHGFS& chGfs_, CHContainer& chV_, CHContainer& chVOld_,
                         NSGFS& nsGfs_, NSContainer& nsVOld_) :
        chLocal(chGfs_,chV_),
        //chOldLocal(chGfs_,chVOld_),
        nsOldLocal(nsGfs_,nsVOld_),
        param(param_)
      {
      }


      template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
      {

        //Quadrature Rule
        const int dim = EG::Entity::dimension;
        const int order = 5; //TODO: Choose more educated
        auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

        chLocal.DoEvaluation(eg);
        //chOldLocal.DoEvaluation(eg);
        nsOldLocal.DoSubEvaluation(eg,child(nsOldLocal.lfs,Dune::Indices::_0));

        for(const auto& ip : rule)
        {
          const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
          auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

          CHVariables_Threephase_Extension<RF,dim,decltype(chLocal.lfs)> ch(chCache, ip.position(), chLocal.lfs, chLocal.lv, S);
          NSVariables_P_VOld_Extension<RF,dim,LFSV> ns(vCache, pCache, ip.position(), lfsv, x.base(), nsOldLocal.lv, S);


          RF pf3=1-ch.pf-ch.pf2;
          RF pf_f=ch.pf+ch.pf2+2*param.delta*pf3;
          Dune::FieldVector<RF,dim> gradpf_f(0.0);
          gradpf_f.axpy((1-2*param.delta),ch.gradpf+ch.gradpf2);

          RF rho_f=param.phys.rho_f(ch.pf,ch.pf2);
          RF rho_f_reg = rho_f + (param.phys.rho_1+param.phys.rho_2)*param.delta;

          for (size_t i=0; i<ns.pspace.size(); i++)
          {
            for(int k=0; k<dim; k++)
            {
              // integrate: (div v)*phi
              //r.accumulate(ns.pspace,i, -1.0* pf_f*(v*gradphi[i][0])*factor);            Weak form
              r.accumulate(ns.pspace,i, -1.0 * (pf_f*ns.div_v  + (gradpf_f*ns.v) )* ns.pbasis.phi[i] * factor);
            }
          }
          for (size_t i=0; i<ns.vfemspace.size(); i++)
          {
            for(int k=0; k<dim; k++)
            {
              // integrate: ( (rho+rhoOld)/2 * v_n+1 - rhoOld * v_n)/dt * phi   + nu * Dv_n+1 : D phi - p_n+1 div phi

              //std::cout << (param.nu * (jacv[k]*gradvphi[i][0]) - p * gradvphi[i][0][k]
              //- param.rho*g[k]*vphi[i] )*factor << std::endl;

              //RF lambda = 3;
              //const auto v_nabla_v = v * jacv[k];
              RF gradxi3k = -param.tw.Sigma3*(ch.gradxi[k]/param.tw.Sigma1 + ch.gradxi2[k]/param.tw.Sigma2);
              RF S = 0;
              //S += xi1*(gradpf1[k]*pf_f - pf1 * gradpf_f[k])/pf_f;
              //S += xi1*(gradpf2[k]*pf_f - pf2 * gradpf_f[k])/pf_f;
              //S += 2*param.delta*(gradxi3k-gradxi2[k]-gradxi1[k]);
              S += ch.pf*ch.gradxi[k] + ch.pf2*ch.gradxi2[k];
              S = S*param.SurfaceTensionScaling*ns.vbasis.phi[i];

              r.accumulate(child(ns.vspace,k),i, (   rho_f_reg*(ns.v[k]-ns.vOld[k])*ns.vbasis.phi[i]/param.time.dt          //TimeEvolution
                                                - ns.p * (pf_f*ns.vbasis.gradphi[i][0][k] + ns.vbasis.phi[i]*gradpf_f[k])            //Pressure
                                                + param.phys.mu * (ns.jacv[k]*ns.vbasis.gradphi[i][0]) //+vphi[i]*(jacv[k]*gradpf)   )  //Viscosity
                                                + param.vDis.eval(pf_f)*ns.v[k]*ns.vbasis.phi[i]              //Velocity Dissipation in Solid
                                                //- rho*g[k]*vphi[i]                //Gravitation
                                                + S                                //Surface Tension
                                              )*factor);

            }
          }
        }
      }

    };
