#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>

//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class FEM_FFS_CHNS_R_CahnHilliard :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<FEM_FFS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<FEM_FFS_CHNS_R_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
{
private:
  typedef typename CHContainer::field_type RF;
  typedef typename CHGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CHLocalBasisCache<CHGFS> chCache;
  VLocalBasisCache<NSGFS> vCache;

  typedef GFS_LocalViewer<CHGFS, CHContainer, ENTITY, RF> CHLView;
  mutable CHLView chOldLocal;

  typedef GFS_LocalViewer<NSGFS, NSContainer, ENTITY, RF> NSLView;
  mutable NSLView nsLocal;


  Param& param; // parameter functions

public:
  enum {doPatternVolume=true};
  //enum {doLambdaVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  FEM_FFS_CHNS_R_CahnHilliard(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, NSGFS& nsGfs_, NSContainer& nsV) :
    chOldLocal(chGfs_,chVOld_),
    nsLocal(nsGfs_,nsV),
    param(param_)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 2; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chOldLocal.DoEvaluation(eg);
    //chOldLocal.DoEvaluation(eg);
    nsLocal.DoSubEvaluation(eg,child(nsLocal.lfs,Dune::Indices::_0));

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHVariables_Threephase_U_PFOld_Extension<RF,dim,LFSV> ch(chCache, ip.position(), lfsv, x.base(),chOldLocal.lv, S);
      NSVariables<RF,dim,decltype(nsLocal.lfs)> ns(vCache, ip.position(), nsLocal.lfs, nsLocal.lv, S);

      RF pf3 = 1-ch.pf-ch.pf2;
      RF xi3 = -param.tw.Sigma3*(ch.xi/param.tw.Sigma1 + ch.xi2/param.tw.Sigma2);

      RF pf_c = ch.pf;
      RF pf_c_Old = ch.pfOld;
      RF pf_c_reg = pf_c+param.delta;
      RF pf_c_reg_Old = pf_c_Old + param.delta;

      for (size_t i=0; i<ch.pfspace.size(); i++)
      {
        RF R1 = 0;//-param.tw.eval_q(ch.pf,ch.pf2,pf3)/param.eps*
                  //(param.reaction.eval(ch.u)+param.phys.curvatureAlpha*(ch.xi-xi3))*phi[i];
        RF J1 = param.mobility.eval(ch.pf,ch.pf2,pf3)*param.eps/param.tw.Sigma1 * (ch.gradxi*ch.basis.gradphi[i][0]);

      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(ch.pfspace,i,factor*(
          (ch.pf-ch.pfOld)*ch.basis.phi[i]/param.time.dt
          +J1
          +(ch.pf*ns.div_v + (ch.gradpf*ns.v))*ch.basis.phi[i]
          //-pf1*(v*gradphi[i][0])   weak transport
          -R1
        ));

        r.accumulate(ch.pf2space,i,factor*(
          (ch.pf2-ch.pf2Old)*ch.basis.phi[i]/param.time.dt
          +param.mobility.eval(ch.pf,ch.pf2,pf3)*param.eps/param.tw.Sigma2 * (ch.gradxi2*ch.basis.gradphi[i][0])
          +(ch.pf2*ns.div_v + (ch.gradpf2*ns.v))*ch.basis.phi[i]
          //-pf2*(v*gradphi[i][0])
        ));

        r.accumulate(ch.xispace,i,factor*(
          ch.xi*ch.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf1(ch.pf,ch.pf2,pf3) *ch.basis.phi[i]
          - param.eps * param.tw.Sigma1* (ch.gradpf*ch.basis.gradphi[i][0])
        ));

        r.accumulate(ch.xi2space,i,factor*(
          ch.xi2*ch.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf2(ch.pf,ch.pf2,pf3) *ch.basis.phi[i]
          - param.eps * param.tw.Sigma2* (ch.gradpf2*ch.basis.gradphi[i][0])
        ));

      // integrate (pf-pfold)*(u-uast)*phi_i/delta_t + pfold*(u-uold)*phi_i/delta_t + D*pf*gradu*grad phi_i = 0
        r.accumulate(ch.uspace,i,factor*(

          //(u-param.uAst)*phi[i]/param.dt //For Testing

          (pf_c_reg-pf_c_reg_Old)*(ch.u-param.phys.uAst)*ch.basis.phi[i]/param.time.dt
          + pf_c_reg_Old*(ch.u-ch.uOld)*ch.basis.phi[i]/param.time.dt
          + param.phys.D*pf_c_reg_Old*(ch.gradu*ch.basis.gradphi[i][0])
          + J1*(ch.u-param.phys.uAst)
          //- pf_c*(u-param.phys.uAst)*(v*gradphi[i][0])
          + (pf_c*(ch.u-param.phys.uAst)*ns.div_v + (ch.u-param.phys.uAst)*(ch.gradpf*ns.v) + pf_c*(ch.gradu*ns.v) )*ch.basis.phi[i]
          ));
      }
    }
  }

};
